/* 
 *  audio.c
 *
 *     Copyright (C) 2000 - Alastair JW Mayer 
 *
 *  This file is part of a DV decoder.  It converts the audio
 *  data stream to playable audio data on a per-frame basis.
 *
 *  audio.c is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your
 *  option) any later version.
 *   
 *  audio.c is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with this file; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 *  Alastair Mayer can be reached at <al@bix.com> or <alastair@ajwm.net>
 *
 */


/*
 *  General commentary:
 *
 *     This collection of routines is intended as part of a larger
 *  DV decoding/playback tool, e.g. libdv.  The functions here
 *  comprise an initialization function called at the beginning of
 *  each frame,  a decoding function called for each audio data DIF
 *  block, and a finalization function called at the end of the
 *  frame.  (The actual chronology may vary -- the user may choose
 *  to collect all the data and call all the routines at the end
 *  of a frame's worth, or distribute the calls as the data is
 *  collected.)
 *
 *  Error checking is currently minimal, it's up to the caller to
 *  make sure these routines are called in the right order.
 *  WARNING:  output of random data may be hazardous to the health
 *  of audio speaker systems.
 *
 *	Usage:
 *		aud_init_sound() -- initializes the sound card device,
 *                        call it once at beginning of playback
 *		aud_init_frame(frametype) -- clears the decoding buffers,
 *                        sets number of samples expected, call at
 *                        the beginning of each frame
 *      aud_decode_block(data, difseq, blocknum) -- shuffles the
 *                        audio samples from the DIF block into 
 *                        the decoding buffer, call for each block
 *      aud_end_frame()  -- outputs the sound samples for the whole
 *                        frame to the sound device.  Tries to
 *                        accomodate some card limitations.  Call
 *                        at the end of each frame.
 */


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/soundcard.h>

#define aud_error(s) \
	fprintf(stderr, "AUD_ERROR: %s\n", s)

#define NTSC	525
#define PAL	625

/*
 * The _50 suffix is for 625/50 (PAL) data,
 * the _60 suffix is for 525/60 (NTSC) data.
 */

/*
 * DV audio data is shuffled within the frame data.
 * Macros and tables here are used to deshuffle.
 */

static int aud_offsets_60[5][9] = {
	{ 0, 15, 30, 10, 25, 40,  5, 20, 35 },
	{ 3, 18, 33, 13, 28, 43,  8, 23, 38 },
	{ 6, 21, 36,  1, 16, 31, 11, 26, 41 },
	{ 9, 24, 39,  4, 19, 34, 14, 29, 44 },
	{12, 27, 42,  7, 22, 37,  2, 17, 32 }
};

/* TODO: fill in this table */
static int aud_offsets_50[6][9];

/*
 * compute the offset of an audio sample
 * ds = DIF sequence number
 * db = DIF block number
 * bp = byte position in block
 * Note: channel 1 data is stored in blocks 0-4,
 *       channel 2 data is stored in blocks 5-9
 *       For channel 2 db = blocknumber - 5
 */

#define aud_samp_offset_60(ds, db, bp) \
	(aud_offsets_60[ds][db] + (bp - 8)/2 * 45)

#define aud_samp_offset_50(ds, db, bp) \
	(aud_offsets_50[ds][db] + (bp - 8)/2 * 54)

/*
 * allocate buffer space for the audio samples per frame
 * NTSC is 1600 or 1602 16-bit samples, PAL is 1920 samples
 */

static short	aud_chan_1[1920];
static short	aud_chan_2[1920];
static int	aud_num_samples;
static int	aud_frame_type;
static short	audiobuf[3840];

static int		aud_format = -1;	/* eg 16-bit, -1 = invalid */
static int		aud_channels;		/* 1=mono, 2=stereo */
static int		aud_rate;		/* sampling rate, ie 48000 */
static int		aud_fd = 0;		/* file descriptor of device */

void aud_init_sound() {
  /* initialize the sound card */
  aud_format = -1;
  
  if ((aud_fd = open("/dev/audio", O_WRONLY, 0)) == -1) {
    perror("/dev/audio, try --nosound");
    aud_format = -1;
    return;
  }
  
  /* set sample format -- try for 16bit */
  aud_format = AFMT_S16_NE;
  if (ioctl(aud_fd, SNDCTL_DSP_SETFMT, &aud_format) == -1) {
    perror("SNDCTL_DSP_SETFMT, try --nosound");
    aud_format = -1;
    close(aud_fd);
    return;
  }
  if (aud_format != AFMT_S16_NE) {
		/* device doesn't support 16 bit, maybe we can do 8 bit? */
    if (aud_format != AFMT_U8) {
      fprintf(stderr, "soundcard doesn't support format\n");
      aud_format = -1;
      close(aud_fd);
      return;
    }
  }
  
  /* set stereo mode */
  aud_channels = 2;
  if (ioctl(aud_fd, SNDCTL_DSP_CHANNELS, &aud_channels) == -1) {
    perror("SNDCTL_DSP_CHANNELS, try --nosound");
    aud_format = -1;
    close(aud_fd);
    return;
  }
  if (aud_channels != 2) {
    fprintf(stderr, "soundcard doesn't support stereo, try --nosound\n");
  }
  
  /* set sampling rate */
  aud_rate = 48000;
  if (ioctl(aud_fd, SNDCTL_DSP_SPEED, &aud_rate) == -1) {
    perror("SNDCTL_DSP_SPEED");
    aud_format = -1;
    close(aud_fd);
    return;
  }
  fprintf(stderr,"audio rate found : %d Hz\n",aud_rate);

  if (abs(48000 - aud_rate) > 4000) {
    /* rate not supported. in future we could try converting */
    fprintf(stderr, "soundcard doesn't support 48kHz (does %d), try --nosound\n", aud_rate);
    aud_format = -1;
    close(aud_fd);
    return;
  }
}
void aud_init_frame(int ftype) {
  int i;
  
  for (i=0; i<1920; i++) {
    aud_chan_1[i] = 0;
    aud_chan_2[i] = 0;
  }
  aud_frame_type = ftype;
  if (ftype == NTSC)
    aud_num_samples = 1600;
  else
    aud_num_samples = 1920;
}

int aud_decode_block(unsigned char* data, int difnum, int blocknum) {
  
  int		bp;
  int		i;
  short	*samp;
  int		afsize;
  
  if ((data[0] & 0xe0) != 0x60) {
    aud_error("Not an audio block.");
    return -1;
  }
  
  if (data[3] == 0x50) { // AS
    afsize = data[4] & 077;
    if (afsize == 0x14) aud_num_samples = 1600;
    else if (afsize == 0x16) aud_num_samples = 1602;
    else if (afsize == 0x18) aud_num_samples = 1920;
  }
  
  if (difnum < 5)
    samp = aud_chan_1;		// channel 1 in first 5 blocks
  else {
    samp = aud_chan_2;
    difnum -= 5;		// subtract 5 for calculating offset
  }
  
  if (aud_frame_type == NTSC) {
    for (bp = 8; bp < 80; bp+=2) {
      i = aud_samp_offset_60(difnum, blocknum, bp);
      samp[i] = ((unsigned char)data[bp] * 256) + (unsigned char)data[bp+1];
    }
  } else {
    for (bp = 8; bp < 80; bp+=2) {
      i = aud_samp_offset_50(difnum, blocknum, bp);
      samp[i] = ((unsigned char)data[bp] * 256) + (unsigned char)data[bp+1];
    }
  } 
  return 0;
}

void aud_end_frame (void) {
  /* we ought to check here that we got a full set of data.
   * also, the fade-in or fade-out could be done by manipulating
   * the data appropriately
   */
  
  /* check aud_format to make sure we initialized the card */
  int i;
  unsigned char* u8buf;
  
  if (aud_format != -1) {	
    if (aud_format == AFMT_S16_NE) {	/* 16 bit */
      if (aud_channels == 2) {		/* stereo */
				/* interleave the two channels */
	for (i=0; i<aud_num_samples; i++) {
	  audiobuf[2*i] = aud_chan_1[i];
	  audiobuf[2*i+1] = aud_chan_2[i];
	}
	/*
	 * the following write() sends the data to the sound card.  If the
	 * overall decoding program is playing back in less than real time,
	 * you either don't want to do this (you'll get wierd audio effects
	 * at the end of each frame -- you could mute the sound I suppose).
	 * Otherwise (or in addition, take out the #else) we write the samples
	 * to a raw audio file ("decoded.dat" here) which can be played back
	 * (also edited, etc) in a program like Broadcast 2000 (2 channel 48khz).
	 */
	write(aud_fd, audiobuf, 2*2*aud_num_samples);
      } else {	
	/* mono */
	/* just do channel 1.  in future we could */
	/* try doing a mixdown here. */
        write(aud_fd, aud_chan_1, 2*aud_num_samples);
      }
    } else if (aud_format == AFMT_U8) {	/* 8 bit unsigned */
      u8buf = (unsigned char*) audiobuf;
      if (aud_channels == 2) {   /* stereo */
	for (i=0; i<aud_num_samples; i++) {
	  u8buf[2*i] = 128+(aud_chan_1[i]>>8);
	  u8buf[2*i+1] = 128+(aud_chan_2[i]>>8);
	}
	write(aud_fd, u8buf, 2*aud_num_samples);
      } else { /* mono */
	for (i=0; i<aud_num_samples; i++) {
	  u8buf[i] = 128+(aud_chan_1[i]>>8);
	}
	write(aud_fd, u8buf, aud_num_samples);
      }
    } /* endif 8 bit */
  }
}

